@@ -17,7 +17,7 @@
 #include <memory>
 #include <sstream>
 
-#include "auth/AuthSupported.h"
+#include "auth/AuthMethodList.h"
 #include "auth/Crypto.h"
 #include "auth/KeyRing.h"
 #include "common/ConfUtils.h"
@@ -34,72 +34,55 @@
 using std::auto_ptr;
 using namespace std;
 
-int KeyRing::from_ceph_context(CephContext *cct, KeyRing **pkeyring)
+int KeyRing::from_ceph_context(CephContext *cct)
 {
   const md_config_t *conf = cct->_conf;
-  bool found_key = false;
-  auto_ptr < KeyRing > keyring(new KeyRing());
 
-  AuthSupported supported(cct);
-
-  if (!supported.is_supported_auth(CEPH_AUTH_CEPHX)) {
-    ldout(cct, 2) << "KeyRing::from_ceph_context: CephX auth is not supported." << dendl;
-    *pkeyring = keyring.release();
-    return 0;
-  }
-
-  int ret = 0;
+  int ret = -ENOENT;
   string filename;
+
   if (ceph_resolve_file_search(conf->keyring, filename)) {
-    ret = keyring->load(cct, filename);
-    if (ret) {
-      lderr(cct) << "KeyRing::from_ceph_context: failed to load " << filename
+    ret = load(cct, filename);
+    if (ret < 0)
+      lderr(cct) << "failed to load " << filename
 		 << ": " << cpp_strerror(ret) << dendl;
-    } else {
-      found_key = true;
-    }
   }
 
   if (!conf->key.empty()) {
     EntityAuth ea;
-    ea.key.decode_base64(conf->key);
-    keyring->add(conf->name, ea);
-    found_key = true;
-  }
-
-  if (!conf->keyfile.empty()) {
-    FILE *fp = fopen(conf->keyfile.c_str(), "r");
-    if (fp) {
-      char buf[100];
-      int res = fread(buf, 1, sizeof(buf) - 1, fp);
-      if (res < 0) {
-	res = ferror(fp);
-	lderr(cct) << "KeyRing::from_ceph_context: failed to read '" << conf->keyfile
-		   << "'" << dendl;
-      }
-      else {
-	string k = buf;
-	EntityAuth ea;
-	ea.key.decode_base64(k);
-	keyring->add(conf->name, ea);
-	found_key = true;
-      }
-      fclose(fp);
-    } else {
-      ret = errno;
-      lderr(cct) << "KeyRing::conf_ceph_context: failed to open " << conf->keyfile
-		 << ": " << cpp_strerror(ret) << dendl;
+    try {
+      ea.key.decode_base64(conf->key);
+      add(conf->name, ea);
+      return 0;
+    }
+    catch (buffer::error& e) {
+      lderr(cct) << "failed to decode key '" << conf->key << "'" << dendl;
+      return -EINVAL;
     }
   }
 
-  if (!found_key) {
-    if (conf->keyring.length())
-      lderr(cct) << "failed to open keyring from " << conf->keyring << dendl;
-    return -ENOENT;
+  if (!conf->keyfile.empty()) {
+    bufferlist bl;
+    string err;
+    int r = bl.read_file(conf->keyfile.c_str(), &err);
+    if (r < 0) {
+      lderr(cct) << err << dendl;
+      return r;
+    }
+    string k(bl.c_str(), bl.length());
+    EntityAuth ea;
+    try {
+      ea.key.decode_base64(k);
+      add(conf->name, ea);
+    }
+    catch (buffer::error& e) {
+      lderr(cct) << "failed to decode key '" << k << "'" << dendl;
+      return -EINVAL;
+    }
+    return 0;
   }
 
-  *pkeyring = keyring.release();
-  return 0;
+  return ret;
 }
 
 KeyRing *KeyRing::create_empty()
@@ -154,7 +137,8 @@ void KeyRing::decode_plaintext(bufferlist::iterator& bli)
   bli.copy_all(bl);
   ConfFile cf;
   std::deque<std::string> parse_errors;
-  if (cf.parse_bufferlist(&bl, &parse_errors) != 0) {
+
+  if (cf.parse_bufferlist(&bl, &parse_errors, NULL) != 0) {
     throw buffer::malformed_input("cannot parse buffer");
   }
 
@@ -252,7 +236,8 @@ void KeyRing::import(CephContext *cct, KeyRing& other)
   for (map<EntityName, EntityAuth>::iterator p = other.keys.begin();
        p != other.keys.end();
        ++p) {
-    ldout(cct, 10) << " importing " << p->first << " " << p->second << dendl;
+    ldout(cct, 10) << " importing " << p->first << dendl;
+    ldout(cct, 30) << "    " << p->second << dendl;
     keys[p->first] = p->second;
   }
 }
