@@ -23,18 +23,17 @@
 #include "include/compat.h"
 
 #include <fcntl.h>
+#include <limits.h>
 #include <sstream>
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
 
-#if defined(__FreeBSD__)
-#include <sys/disk.h>
-#endif
+#include "common/blkdev.h"
 
 
-#define DOUT_SUBSYS journal
+#define dout_subsys ceph_subsys_journal
 #undef dout_prefix
 #define dout_prefix *_dout << "journal "
 
@@ -58,7 +57,7 @@ int FileJournal::_open(bool forwrite, bool create)
   if (forwrite) {
     flags = O_RDWR;
     if (directio)
-      flags |= O_DIRECT | O_SYNC;
+      flags |= O_DIRECT | O_DSYNC;
   } else {
     flags = O_RDONLY;
   }
@@ -75,8 +74,8 @@ int FileJournal::_open(bool forwrite, bool create)
   fd = TEMP_FAILURE_RETRY(::open(fn.c_str(), flags, 0644));
   if (fd < 0) {
     int err = errno;
-    derr << "FileJournal::_open: unable to open journal: open() failed: "
-	 << cpp_strerror(err) << dendl;
+    dout(2) << "FileJournal::_open: unable to open journal: open() failed: "
+	    << cpp_strerror(err) << dendl;
     return -err;
   }
 
@@ -124,23 +123,8 @@ int FileJournal::_open(bool forwrite, bool create)
 
 int FileJournal::_open_block_device()
 {
-  int ret = 0;
   int64_t bdev_sz = 0;
-#if defined(__FreeBSD__)
-  ret = ::ioctl(fd, DIOCGMEDIASIZE, &bdev_sz);
-#elif defined(__linux__)
-#ifdef BLKGETSIZE64
-  // ioctl block device
-  ret = ::ioctl(fd, BLKGETSIZE64, &bdev_sz);
-#elif BLKGETSIZE
-  // hrm, try the 32 bit ioctl?
-  unsigned long sectors = 0;
-  ret = ::ioctl(fd, BLKGETSIZE, &sectors);
-  bdev_sz = sectors * 512ULL;
-#endif
-#else
-#error "Compile error: we don't know how to get the size of a raw block device."
-#endif /* !__FreeBSD__ */
+  int ret = get_block_device_size(fd, &bdev_sz);
   if (ret) {
     dout(0) << __func__ << ": failed to read block device size." << dendl;
     return -EIO;
@@ -155,23 +139,11 @@ int FileJournal::_open_block_device()
 
   int64_t conf_journal_sz(g_conf->osd_journal_size);
   conf_journal_sz <<= 20;
-  if (bdev_sz < conf_journal_sz) {
-    dout(0) << __func__ << ": you have configured a journal of size "
-      << conf_journal_sz << ", but the block device " << fn << " is only "
-      << bdev_sz << " bytes in size." << dendl;
-    return -EINVAL;
-  }
 
-  if (conf_journal_sz == 0) {
-    dout(10) << __func__ << ": no journal size specified in configuration. "
-      << "We'll use the entire block device (size: " << bdev_sz << ")" << dendl;
-    max_size = bdev_sz;
-  }
-  else {
-    dout(10) << __func__ << ": using " << conf_journal_sz << " bytes "
-             << "of a " << bdev_sz << " byte block device." << dendl;
-    max_size = conf_journal_sz;
-  }
+  dout(10) << __func__ << ": ignoring osd journal size. "
+	   << "We'll use the entire block device (size: " << bdev_sz << ")"
+	   << dendl;
+  max_size = bdev_sz;
 
   /* block devices have to write in blocks of CEPH_PAGE_SIZE */
   block_size = CEPH_PAGE_SIZE;
@@ -334,12 +306,65 @@ int FileJournal::_open_file(int64_t oldsize, blksize_t blksize,
   }
   block_size = MAX(blksize, (blksize_t)CEPH_PAGE_SIZE);
 
+  if (create && g_conf->journal_zero_on_create) {
+    derr << "FileJournal::_open_file : zeroing journal" << dendl;
+    uint64_t write_size = 1 << 20;
+    char *buf = new char[write_size];
+    memset(static_cast<void*>(buf), 0, write_size);
+    uint64_t i = 0;
+    for (; (i + write_size) <= (unsigned)max_size; i += write_size) {
+      ret = ::pwrite(fd, static_cast<void*>(buf), write_size, i);
+      if (ret < 0) {
+	delete [] buf;
+	return -errno;
+      }
+    }
+    if (i < (unsigned)max_size) {
+      ret = ::pwrite(fd, static_cast<void*>(buf), max_size - i, i);
+      if (ret < 0) {
+	delete [] buf;
+	return -errno;
+      }
+    }
+    delete [] buf;
+  }
+      
+
   dout(10) << "_open journal is not a block device, NOT checking disk "
            << "write cache on '" << fn << "'" << dendl;
 
   return 0;
 }
 
+int FileJournal::check()
+{
+  int ret;
+
+  ret = _open(false, false);
+  if (ret < 0)
+    goto done;
+
+  ret = read_header();
+  if (ret < 0)
+    goto done;
+
+  if (header.fsid != fsid) {
+    derr << "check: ondisk fsid " << header.fsid << " doesn't match expected " << fsid
+	 << ", invalid (someone else's?) journal" << dendl;
+    ret = -EINVAL;
+    goto done;
+  }
+
+  dout(1) << "check: header looks ok" << dendl;
+  ret = 0;
+
+ done:
+  TEMP_FAILURE_RETRY(::close(fd));
+  fd = -1;
+  return ret;
+}
+
+
 int FileJournal::create()
 {
   void *buf = 0;
@@ -393,7 +418,7 @@ int FileJournal::create()
     goto free_buf;
   }
 
-  needed_space = g_conf->osd_max_write_size << 20;
+  needed_space = ((int64_t)g_conf->osd_max_write_size) << 20;
   needed_space += (2 * sizeof(entry_header_t)) + get_top();
   if (header.max_size - header.start < needed_space) {
     derr << "FileJournal::create: OSD journal is not large enough to hold "
@@ -530,7 +555,7 @@ void FileJournal::close()
   stop_writer();
 
   // close
-  assert(writeq.empty());
+  assert(writeq_empty());
   assert(fd >= 0);
   TEMP_FAILURE_RETRY(::close(fd));
   fd = -1;
@@ -548,6 +573,9 @@ int FileJournal::dump(ostream& out)
 
   read_pos = header.start;
 
+  JSONFormatter f(true);
+
+  f.open_array_section("journal");
   while (1) {
     bufferlist bl;
     uint64_t seq = 0;
@@ -557,16 +585,27 @@ int FileJournal::dump(ostream& out)
       break;
     }
 
-    out << "offset " << pos << " seq " << seq << "\n";
-   
+    f.open_object_section("entry");
+    f.dump_unsigned("offset", pos);
+    f.dump_unsigned("seq", seq);
+    f.open_array_section("transactions");
     bufferlist::iterator p = bl.begin();
+    int trans_num = 0;
     while (!p.end()) {
       ObjectStore::Transaction *t = new ObjectStore::Transaction(p);
-      t->dump(out);
+      f.open_object_section("transaction");
+      f.dump_unsigned("trans_num", trans_num);
+      t->dump(&f);
+      f.close_section();
       delete t;
+      trans_num++;
     }
+    f.close_section();
+    f.close_section();
+    f.flush(cout);
   }
-  out << std::endl;
+
+  f.close_section();
   dout(10) << "dump finish" << dendl;
   return 0;
 }
@@ -583,15 +622,19 @@ void FileJournal::start_writer()
 
 void FileJournal::stop_writer()
 {
-  write_lock.Lock();
   {
-    write_stop = true;
-    write_cond.Signal();
+    Mutex::Locker l(write_lock);
 #ifdef HAVE_LIBAIO
+    Mutex::Locker q(aio_lock);
+#endif
+    Mutex::Locker p(writeq_lock);
+    write_stop = true;
+    writeq_cond.Signal();
+#ifdef HAVE_LIBAIO
+    aio_cond.Signal();
     write_finish_cond.Signal();
 #endif
   } 
-  write_lock.Unlock();
   write_thread.join();
 #ifdef HAVE_LIBAIO
   write_finish_thread.join();
@@ -683,7 +726,7 @@ int FileJournal::check_for_full(uint64_t seq, off64_t pos, off64_t size)
     if (room < (header.max_size >> 1) &&
 	room + size > (header.max_size >> 1)) {
       dout(10) << " passing half full mark, triggering commit" << dendl;
-      do_sync_cond->Signal();  // initiate a real commit so we can trim
+      do_sync_cond->SloppySignal();  // initiate a real commit so we can trim
     }
   }
 
@@ -702,7 +745,7 @@ int FileJournal::check_for_full(uint64_t seq, off64_t pos, off64_t size)
 
   off64_t max = header.max_size - get_top();
   if (size > max)
-    dout(0) << "JOURNAL TOO SMALL: item " << size << " > journal " << max << " (usable)" << dendl;
+    dout(0) << "JOURNAL TOO SMALL: continuing, but slow: item " << size << " > journal " << max << " (usable)" << dendl;
 
   return -ENOSPC;
 }
@@ -718,7 +761,7 @@ int FileJournal::prepare_multi_write(bufferlist& bl, uint64_t& orig_ops, uint64_
   if (full_state != FULL_NOTFULL)
     return -ENOSPC;
   
-  while (!writeq.empty()) {
+  while (!writeq_empty()) {
     int r = prepare_single_write(bl, queue_pos, orig_ops, orig_bytes);
     if (r == -ENOSPC) {
       if (orig_ops)
@@ -734,9 +777,9 @@ int FileJournal::prepare_multi_write(bufferlist& bl, uint64_t& orig_ops, uint64_
 
 	// throw out what we have so far
 	full_state = FULL_FULL;
-	while (!writeq.empty()) {
-	  put_throttle(1, writeq.front().bl.length());
-	  writeq.pop_front();
+	while (!writeq_empty()) {
+	  put_throttle(1, peek_write().bl.length());
+	  pop_write();
 	}  
 	print_header();
       }
@@ -784,33 +827,40 @@ void FileJournal::queue_write_fin(uint64_t seq, Context *fin)
 
 void FileJournal::queue_completions_thru(uint64_t seq)
 {
+  assert(finisher_lock.is_locked());
   utime_t now = ceph_clock_now(g_ceph_context);
-  while (!completions.empty() &&
-	 completions.front().seq <= seq) {
+  while (!completions_empty()) {
+    completion_item next = completion_peek_front();
+    if (next.seq > seq)
+      break;
+    completion_pop_front();
     utime_t lat = now;
-    lat -= completions.front().start;
+    lat -= next.start;
     dout(10) << "queue_completions_thru seq " << seq
-	     << " queueing seq " << completions.front().seq
-	     << " " << completions.front().finish
+	     << " queueing seq " << next.seq
+	     << " " << next.finish
 	     << " lat " << lat << dendl;
     if (logger) {
-      logger->finc(l_os_j_lat, lat);
+      logger->tinc(l_os_j_lat, lat);
     }
-    if (completions.front().finish)
-      finisher->queue(completions.front().finish);
-    completions.pop_front();
+    if (next.finish)
+      finisher->queue(next.finish);
+    if (next.tracked_op)
+      next.tracked_op->mark_event("journaled_completion_queued");
   }
+  finisher_cond.Signal();
 }
 
 int FileJournal::prepare_single_write(bufferlist& bl, off64_t& queue_pos, uint64_t& orig_ops, uint64_t& orig_bytes)
 {
   // grab next item
-  uint64_t seq = writeq.front().seq;
-  bufferlist &ebl = writeq.front().bl;
+  write_item &next_write = peek_write();
+  uint64_t seq = next_write.seq;
+  bufferlist &ebl = next_write.bl;
   unsigned head_size = sizeof(entry_header_t);
   off64_t base_size = 2*head_size + ebl.length();
 
-  int alignment = writeq.front().alignment; // we want to start ebl with this alignment
+  int alignment = next_write.alignment; // we want to start ebl with this alignment
   unsigned pre_pad = 0;
   if (alignment >= 0)
     pre_pad = ((unsigned int)alignment - (unsigned int)head_size) & ~CEPH_PAGE_MASK;
@@ -855,8 +905,11 @@ int FileJournal::prepare_single_write(bufferlist& bl, off64_t& queue_pos, uint64
   }
   bl.append((const char*)&h, sizeof(h));
 
+  if (next_write.tracked_op)
+    next_write.tracked_op->mark_event("write_thread_in_journal_buffer");
+
   // pop from writeq
-  writeq.pop_front();
+  pop_write();
   journalq.push_back(pair<uint64_t,off64_t>(seq, queue_pos));
   writing_seq = seq;
 
@@ -884,9 +937,15 @@ void FileJournal::align_bl(off64_t pos, bufferlist& bl)
 int FileJournal::write_bl(off64_t& pos, bufferlist& bl)
 {
   align_bl(pos, bl);
+  int ret;
 
-  ::lseek64(fd, pos, SEEK_SET);
-  int ret = bl.write_fd(fd);
+  off64_t spos = ::lseek64(fd, pos, SEEK_SET);
+  if (spos < 0) {
+    ret = -errno;
+    derr << "FileJournal::write_bl : lseek64 failed " << cpp_strerror(ret) << dendl;
+    return ret;
+  }
+  ret = bl.write_fd(fd);
   if (ret) {
     derr << "FileJournal::write_bl : write_fd failed: " << cpp_strerror(ret) << dendl;
     return ret;
@@ -907,8 +966,6 @@ void FileJournal::do_write(bufferlist& bl)
     hbp = prepare_header();
   }
 
-  writing = true;
-
   write_lock.Unlock();
 
   dout(15) << "do_write writing " << write_pos << "~" << bl.length() 
@@ -997,65 +1054,66 @@ void FileJournal::do_write(bufferlist& bl)
 
   write_lock.Lock();    
 
-  writing = false;
-  write_empty_cond.Signal();
-
   // wrap if we hit the end of the journal
   if (pos == header.max_size)
     pos = get_top();
   write_pos = pos;
   assert(write_pos % header.alignment == 0);
 
-  journaled_seq = writing_seq;
+  {
+    Mutex::Locker locker(finisher_lock);
+    journaled_seq = writing_seq;
 
-  // kick finisher?  
-  //  only if we haven't filled up recently!
-  if (full_state != FULL_NOTFULL) {
-    dout(10) << "do_write NOT queueing finisher seq " << journaled_seq
-	     << ", full_commit_seq|full_restart_seq" << dendl;
-  } else {
-    if (plug_journal_completions) {
-      dout(20) << "do_write NOT queueing finishers through seq " << journaled_seq
-	       << " due to completion plug" << dendl;
+    // kick finisher?
+    //  only if we haven't filled up recently!
+    if (full_state != FULL_NOTFULL) {
+      dout(10) << "do_write NOT queueing finisher seq " << journaled_seq
+	       << ", full_commit_seq|full_restart_seq" << dendl;
     } else {
-      dout(20) << "do_write queueing finishers through seq " << journaled_seq << dendl;
-      queue_completions_thru(journaled_seq);
+      if (plug_journal_completions) {
+	dout(20) << "do_write NOT queueing finishers through seq " << journaled_seq
+		 << " due to completion plug" << dendl;
+      } else {
+	dout(20) << "do_write queueing finishers through seq " << journaled_seq << dendl;
+	queue_completions_thru(journaled_seq);
+      }
     }
   }
 }
 
 void FileJournal::flush()
 {
-  write_lock.Lock();
-  while ((!writeq.empty() || writing)) {
-    dout(5) << "flush waiting for writeq to empty and writes to complete" << dendl;
-    write_empty_cond.Wait(write_lock);
+  dout(10) << "waiting for completions to empty" << dendl;
+  {
+    Mutex::Locker l(finisher_lock);
+    while (!completions_empty())
+      finisher_cond.Wait(finisher_lock);
   }
-  write_lock.Unlock();
-  dout(5) << "flush waiting for finisher" << dendl;
+  dout(10) << "flush waiting for finisher" << dendl;
   finisher->wait_for_empty();
-  dout(5) << "flush done" << dendl;
+  dout(10) << "flush done" << dendl;
 }
 
 
 void FileJournal::write_thread_entry()
 {
   dout(10) << "write_thread_entry start" << dendl;
-  write_lock.Lock();
-  
-  while (!write_stop || 
-	 !writeq.empty()) {  // ensure we fully flush the writeq before stopping
-    if (writeq.empty()) {
-      // sleep
-      dout(20) << "write_thread_entry going to sleep" << dendl;
-      write_empty_cond.Signal();
-      write_cond.Wait(write_lock);
-      dout(20) << "write_thread_entry woke up" << dendl;
-      continue;
+  while (1) {
+    {
+      Mutex::Locker locker(writeq_lock);
+      if (writeq.empty()) {
+	if (write_stop)
+	  break;
+	dout(20) << "write_thread_entry going to sleep" << dendl;
+	writeq_cond.Wait(writeq_lock);
+	dout(20) << "write_thread_entry woke up" << dendl;
+	continue;
+      }
     }
     
 #ifdef HAVE_LIBAIO
     if (aio) {
+      Mutex::Locker locker(aio_lock);
       // should we back off to limit aios in flight?  try to do this
       // adaptively so that we submit larger aios once we have lots of
       // them in flight.
@@ -1069,13 +1127,14 @@ void FileJournal::write_thread_entry()
 	dout(20) << "write_thread_entry deferring until more aios complete: "
 		 << aio_num << " aios with " << aio_bytes << " bytes needs " << min_new
 		 << " bytes to start a new aio (currently " << cur << " pending)" << dendl;
-	write_cond.Wait(write_lock);
+	aio_cond.Wait(aio_lock);
 	dout(20) << "write_thread_entry woke up" << dendl;
 	continue;
       }
     }
 #endif
 
+    Mutex::Locker locker(write_lock);
     uint64_t orig_ops = 0;
     uint64_t orig_bytes = 0;
 
@@ -1089,6 +1148,11 @@ void FileJournal::write_thread_entry()
     }
     assert(r == 0);
 
+    if (logger) {
+      logger->inc(l_os_j_wr);
+      logger->inc(l_os_j_wr_bytes, bl.length());
+    }
+
 #ifdef HAVE_LIBAIO
     if (aio)
       do_aio_write(bl);
@@ -1097,11 +1161,9 @@ void FileJournal::write_thread_entry()
 #else
     do_write(bl);
 #endif
-    
     put_throttle(orig_ops, orig_bytes);
   }
-  write_empty_cond.Signal();
-  write_lock.Unlock();
+
   dout(10) << "write_thread_entry finish" << dendl;
 }
 
@@ -1118,11 +1180,6 @@ void FileJournal::do_aio_write(bufferlist& bl)
     hbp = prepare_header();
   }
 
-  if (!writing) {
-    writing = true;
-    write_finish_cond.Signal();
-  }
-
   // entry
   off64_t pos = write_pos;
 
@@ -1190,44 +1247,57 @@ void FileJournal::do_aio_write(bufferlist& bl)
  */
 int FileJournal::write_aio_bl(off64_t& pos, bufferlist& bl, uint64_t seq)
 {
+  Mutex::Locker locker(aio_lock);
   align_bl(pos, bl);
 
   dout(20) << "write_aio_bl " << pos << "~" << bl.length() << " seq " << seq << dendl;
   
-  aio_queue.push_back(aio_info(bl, pos, seq));
-  aio_info& aio = aio_queue.back();
-
-  aio.iov = new iovec[aio.bl.buffers().size()];
-  int n = 0;
-  for (std::list<buffer::ptr>::const_iterator p = aio.bl.buffers().begin(); 
-       p != aio.bl.buffers().end();
-       ++p, ++n) {
-    aio.iov[n].iov_base = (void *)p->c_str();
-    aio.iov[n].iov_len = p->length();
-  }
-  io_prep_pwritev(&aio.iocb, fd, aio.iov, n, pos);
-
-  dout(20) << "write_aio_bl .. " << aio.off << "~" << aio.len
-	   << " in " << n << dendl;
-
-  aio_num++;
-  aio_bytes += aio.len;
-
-  iocb *piocb = &aio.iocb;
-  int attempts = 10;
-  do {
-    int r = io_submit(aio_ctx, 1, &piocb);
-    if (r < 0) {
-      derr << "io_submit to " << aio.off << "~" << aio.len
-	   << " got " << cpp_strerror(r) << dendl;
-      if (r == -EAGAIN && attempts-- > 0) {
-	usleep(500);
-	continue;
-      }
-      assert(0 == "io_submit got unexpected error");
+  while (bl.length() > 0) {
+    int max = MIN(bl.buffers().size(), IOV_MAX-1);
+    iovec *iov = new iovec[max];
+    int n = 0;
+    unsigned len = 0;
+    for (std::list<buffer::ptr>::const_iterator p = bl.buffers().begin();
+	 n < max;
+	 ++p, ++n) {
+      assert(p != bl.buffers().end());
+      iov[n].iov_base = (void *)p->c_str();
+      iov[n].iov_len = p->length();
+      len += p->length();
     }
-  } while (false);
-  pos += aio.len;
+
+    bufferlist tbl;
+    bl.splice(0, len, &tbl);  // move bytes from bl -> tbl
+
+    aio_queue.push_back(aio_info(tbl, pos, bl.length() > 0 ? 0 : seq));
+    aio_info& aio = aio_queue.back();
+    aio.iov = iov;
+
+    io_prep_pwritev(&aio.iocb, fd, aio.iov, n, pos);
+
+    dout(20) << "write_aio_bl .. " << aio.off << "~" << aio.len
+	     << " in " << n << dendl;
+
+    aio_num++;
+    aio_bytes += aio.len;
+
+    iocb *piocb = &aio.iocb;
+    int attempts = 10;
+    do {
+      int r = io_submit(aio_ctx, 1, &piocb);
+      if (r < 0) {
+	derr << "io_submit to " << aio.off << "~" << aio.len
+	     << " got " << cpp_strerror(r) << dendl;
+	if (r == -EAGAIN && attempts-- > 0) {
+	  usleep(500);
+	  continue;
+	}
+	assert(0 == "io_submit got unexpected error");
+      }
+    } while (false);
+    pos += aio.len;
+  }
+  write_finish_cond.Signal();
   return 0;
 }
 #endif
@@ -1236,47 +1306,46 @@ void FileJournal::write_finish_thread_entry()
 {
 #ifdef HAVE_LIBAIO
   dout(10) << "write_finish_thread_entry enter" << dendl;
-  write_lock.Lock();
   while (true) {
-    if (aio_queue.empty()) {
-      if (write_stop)
-	break;
-      dout(20) << "write_finish_thread_entry sleeping" << dendl;
-      write_finish_cond.Wait(write_lock);
-      continue;
+    {
+      Mutex::Locker locker(aio_lock);
+      if (aio_queue.empty()) {
+	if (write_stop)
+	  break;
+	dout(20) << "write_finish_thread_entry sleeping" << dendl;
+	write_finish_cond.Wait(aio_lock);
+	continue;
+      }
     }
     
-    write_lock.Unlock();
-
     dout(20) << "write_finish_thread_entry waiting for aio(s)" << dendl;
     io_event event[16];
     int r = io_getevents(aio_ctx, 1, 16, event, NULL);
     if (r < 0) {
       if (r == -EINTR) {
 	dout(0) << "io_getevents got " << cpp_strerror(r) << dendl;
-	write_lock.Lock();
 	continue;
       }
       derr << "io_getevents got " << cpp_strerror(r) << dendl;
       assert(0 == "got unexpected error from io_getevents");
     }
-
-    write_lock.Lock();
-
-    for (int i=0; i<r; i++) {
-      aio_info *ai = (aio_info *)event[i].obj;
-      if (event[i].res != ai->len) {
-	derr << "aio to " << ai->off << "~" << ai->len
-	     << " got " << cpp_strerror(event[i].res) << dendl;
-	assert(0 == "unexpected aio error");
+    
+    {
+      Mutex::Locker locker(aio_lock);
+      for (int i=0; i<r; i++) {
+	aio_info *ai = (aio_info *)event[i].obj;
+	if (event[i].res != ai->len) {
+	  derr << "aio to " << ai->off << "~" << ai->len
+	       << " got " << cpp_strerror(event[i].res) << dendl;
+	  assert(0 == "unexpected aio error");
+	}
+	dout(10) << "write_finish_thread_entry aio " << ai->off
+		 << "~" << ai->len << " done" << dendl;
+	ai->done = true;
       }
-      dout(10) << "write_finish_thread_entry aio " << ai->off << "~" << ai->len << " done" << dendl;
-      ai->done = true;
+      check_aio_completion();
     }
-
-    check_aio_completion();
-  }	 
-  write_lock.Unlock();
+  }
   dout(10) << "write_finish_thread_entry exit" << dendl;
 #endif
 }
@@ -1287,17 +1356,18 @@ void FileJournal::write_finish_thread_entry()
  */
 void FileJournal::check_aio_completion()
 {
-  assert(write_lock.is_locked());
+  assert(aio_lock.is_locked());
   dout(20) << "check_aio_completion" << dendl;
 
   bool completed_something = false;
+  uint64_t new_journaled_seq = 0;
 
   list<aio_info>::iterator p = aio_queue.begin();
   while (p != aio_queue.end() && p->done) {
     dout(20) << "check_aio_completion completed seq " << p->seq << " "
 	     << p->off << "~" << p->len << dendl;
     if (p->seq) {
-      journaled_seq = p->seq;
+      new_journaled_seq = p->seq;
       completed_something = true;
     }
     aio_num--;
@@ -1308,6 +1378,8 @@ void FileJournal::check_aio_completion()
   if (completed_something) {
     // kick finisher?  
     //  only if we haven't filled up recently!
+    Mutex::Locker locker(finisher_lock);
+    journaled_seq = new_journaled_seq;
     if (full_state != FULL_NOTFULL) {
       dout(10) << "check_aio_completion NOT queueing finisher seq " << journaled_seq
 	       << ", full_commit_seq|full_restart_seq" << dendl;
@@ -1322,51 +1394,61 @@ void FileJournal::check_aio_completion()
     }
 
     // maybe write queue was waiting for aio count to drop?
-    if (!writeq.empty())
-      write_cond.Signal();
-
-    // wake up flush?
-    if (aio_queue.empty()) {
-      writing = false;
-      write_empty_cond.Signal();
-      assert(aio_num == 0);
-      assert(aio_bytes == 0);
-    }
+    aio_cond.Signal();
   }
 }
 #endif
 
-void FileJournal::submit_entry(uint64_t seq, bufferlist& e, int alignment, Context *oncommit)
+void FileJournal::submit_entry(uint64_t seq, bufferlist& e, int alignment,
+			       Context *oncommit, TrackedOpRef osd_op)
 {
-  Mutex::Locker locker(write_lock);  // ** lock **
-
   // dump on queue
   dout(5) << "submit_entry seq " << seq
-	   << " len " << e.length()
-	   << " (" << oncommit << ")" << dendl;
+	  << " len " << e.length()
+	  << " (" << oncommit << ")" << dendl;
   assert(e.length() > 0);
 
-  completions.push_back(completion_item(seq, oncommit, ceph_clock_now(g_ceph_context)));
-
-  if (full_state == FULL_NOTFULL) {
-    // queue and kick writer thread
-    dout(30) << "XXX throttle take " << e.length() << dendl;
-    throttle_ops.take(1);
-    throttle_bytes.take(e.length());
-
-    if (logger) {
-      logger->set(l_os_jq_max_ops, throttle_ops.get_max());
-      logger->set(l_os_jq_max_bytes, throttle_bytes.get_max());
-      logger->set(l_os_jq_ops, throttle_ops.get_current());
-      logger->set(l_os_jq_bytes, throttle_bytes.get_current());
-    }
-
-    writeq.push_back(write_item(seq, e, alignment));
-    write_cond.Signal();
-  } else {
-    // not journaling this.  restart writing no sooner than seq + 1.
-    dout(10) << " journal is/was full" << dendl;
+  dout(30) << "XXX throttle take " << e.length() << dendl;
+  throttle_ops.take(1);
+  throttle_bytes.take(e.length());
+  if (osd_op)
+    osd_op->mark_event("commit_queued_for_journal_write");
+  if (logger) {
+    logger->set(l_os_jq_max_ops, throttle_ops.get_max());
+    logger->set(l_os_jq_max_bytes, throttle_bytes.get_max());
+    logger->set(l_os_jq_ops, throttle_ops.get_current());
+    logger->set(l_os_jq_bytes, throttle_bytes.get_current());
   }
+
+  {
+    Mutex::Locker l1(writeq_lock);  // ** lock **
+    Mutex::Locker l2(completions_lock);  // ** lock **
+    completions.push_back(
+      completion_item(
+	seq, oncommit, ceph_clock_now(g_ceph_context), osd_op));
+    writeq.push_back(write_item(seq, e, alignment, osd_op));
+    writeq_cond.Signal();
+  }
+}
+
+bool FileJournal::writeq_empty()
+{
+  Mutex::Locker locker(writeq_lock);
+  return writeq.empty();
+}
+
+FileJournal::write_item &FileJournal::peek_write()
+{
+  assert(write_lock.is_locked());
+  Mutex::Locker locker(writeq_lock);
+  return writeq.front();
+}
+
+void FileJournal::pop_write()
+{
+  assert(write_lock.is_locked());
+  Mutex::Locker locker(writeq_lock);
+  writeq.pop_front();
 }
 
 void FileJournal::commit_start()
@@ -1420,21 +1502,24 @@ void FileJournal::committed_thru(uint64_t seq)
   must_write_header = true;
   print_header();
 
-  // completions!
-  queue_completions_thru(seq);
-  if (plug_journal_completions) {
-    dout(10) << " removing completion plug, queuing completions thru journaled_seq " << journaled_seq << dendl;
-    plug_journal_completions = false;
-    queue_completions_thru(journaled_seq);
+  {
+    Mutex::Locker locker(finisher_lock);
+    // completions!
+    queue_completions_thru(seq);
+    if (plug_journal_completions) {
+      dout(10) << " removing completion plug, queuing completions thru journaled_seq " << journaled_seq << dendl;
+      plug_journal_completions = false;
+      queue_completions_thru(journaled_seq);
+    }
   }
 
   // committed but unjournaled items
-  while (!writeq.empty() && writeq.front().seq <= seq) {
-    dout(15) << " dropping committed but unwritten seq " << writeq.front().seq 
-	     << " len " << writeq.front().bl.length()
+  while (!writeq_empty() && peek_write().seq <= seq) {
+    dout(15) << " dropping committed but unwritten seq " << peek_write().seq 
+	     << " len " << peek_write().bl.length()
 	     << dendl;
-    put_throttle(1, writeq.front().bl.length());
-    writeq.pop_front();  
+    put_throttle(1, peek_write().bl.length());
+    pop_write();
   }
   
   commit_cond.Signal();
@@ -1489,10 +1574,11 @@ void FileJournal::wrap_read_bl(off64_t& pos, int64_t olen, bufferlist& bl)
       len = olen;                         // rest
     
 #ifdef DARWIN
-    ::lseek(fd, pos, SEEK_SET);
+    int64_t actual = ::lseek(fd, pos, SEEK_SET);
 #else
-    ::lseek64(fd, pos, SEEK_SET);
+    int64_t actual = ::lseek64(fd, pos, SEEK_SET);
 #endif
+    assert(actual == pos);
     
     bufferptr bp = buffer::create(len);
     int r = safe_read_exact(fd, bp.c_str(), len);
