@@ -180,6 +180,11 @@ public:
       index();
     }
 
+    void split_into(
+      pg_t child_pgid,
+      unsigned split_bits,
+      IndexedLog *olog);
+
     void zero() {
       unindex();
       pg_log_t::clear();
@@ -281,6 +286,19 @@ public:
     uint64_t zero_to;                // first non-zeroed byte of log.
     bool has_checksums;
 
+    /**
+     * We reconstruct the missing set by comparing the recorded log against
+     * the objects in the pg collection.  Unfortunately, it's possible to
+     * have an object in the missing set which is not in the log due to
+     * a divergent operation with a prior_version pointing before the
+     * pg log tail.  To deal with this, we store alongside the log a mapping
+     * of divergent priors to be checked along with the log during read_state.
+     */
+    map<eversion_t, hobject_t> divergent_priors;
+    void add_divergent_prior(eversion_t version, hobject_t obj) {
+      divergent_priors.insert(make_pair(version, obj));
+    }
+
     OndiskLog() : tail(0), head(0), zero_to(0),
 		  has_checksums(true) {}
 
@@ -294,10 +312,11 @@ public:
     }
 
     void encode(bufferlist& bl) const {
-      ENCODE_START(4, 3, bl);
+      ENCODE_START(5, 3, bl);
       ::encode(tail, bl);
       ::encode(head, bl);
       ::encode(zero_to, bl);
+      ::encode(divergent_priors, bl);
       ENCODE_FINISH(bl);
     }
     void decode(bufferlist::iterator& bl) {
@@ -309,6 +328,8 @@ public:
 	::decode(zero_to, bl);
       else
 	zero_to = 0;
+      if (struct_v >= 5)
+	::decode(divergent_priors, bl);
       DECODE_FINISH(bl);
     }
     void dump(Formatter *f) const {
@@ -348,7 +369,6 @@ protected:
    * put_unlock() when done with the current pointer (_most common_).
    */  
   Mutex _lock;
-  Mutex _qlock;
   Cond _cond;
   atomic_t ref;
 
@@ -356,18 +376,12 @@ public:
   bool deleting;  // true while RemoveWQ should be chewing on us
 
   void lock(bool no_lockdep = false);
-  void lockq(bool no_lockdep = false) {
-    _qlock.Lock(no_lockdep);
-  }
   void unlock() {
     //generic_dout(0) << this << " " << info.pgid << " unlock" << dendl;
     assert(!dirty_info);
     assert(!dirty_log);
     _lock.Unlock();
   }
-  void unlockq() {
-    _qlock.Unlock();
-  }
 
   /* During handle_osd_map, the osd holds a write lock to the osdmap.
    * *_with_map_lock_held assume that the map_lock is already held */
@@ -403,8 +417,6 @@ public:
   }
 
 
-  list<OpRequestRef> op_queue;  // op queue
-
   bool dirty_info, dirty_log;
 
 public:
@@ -644,8 +656,11 @@ protected:
   list<OpRequestRef>            waiting_for_all_missing;
   map<hobject_t, list<OpRequestRef> > waiting_for_missing_object,
                                         waiting_for_degraded_object;
+  // Callbacks should assume pg (and nothing else) is locked
+  map<hobject_t, list<Context*> > callbacks_for_degraded_object;
   map<eversion_t,list<OpRequestRef> > waiting_for_ack, waiting_for_ondisk;
   map<eversion_t,OpRequestRef>   replay_queue;
+  void split_ops(PG *child, unsigned split_bits);
 
   void requeue_object_waiters(map<hobject_t, list<OpRequestRef> >& m);
   void requeue_ops(list<OpRequestRef> &l);
@@ -677,6 +692,7 @@ public:
   }
   
   bool needs_recovery() const;
+  bool needs_backfill() const;
 
   void mark_clean();  ///< mark an active pg clean
 
@@ -749,7 +765,7 @@ public:
 		list<Context*>& tfin,
 		map< int, map<pg_t,pg_query_t> >& query_map,
 		map<int, vector<pair<pg_notify_t, pg_interval_map_t> > > *activator_map=0);
-  void _activate_committed(epoch_t e, entity_inst_t& primary);
+  void _activate_committed(epoch_t e);
   void all_activated_and_committed();
 
   void proc_primary_info(ObjectStore::Transaction &t, const pg_info_t &info);
@@ -771,7 +787,7 @@ public:
 
   Context *finish_sync_event;
 
-  void finish_recovery(ObjectStore::Transaction& t, list<Context*>& tfin);
+  void finish_recovery(list<Context*>& tfin);
   void _finish_recovery(Context *c);
   void cancel_recovery();
   void clear_recovery_state();
@@ -781,6 +797,9 @@ public:
   void start_recovery_op(const hobject_t& soid);
   void finish_recovery_op(const hobject_t& soid, bool dequeue=false);
 
+  void split_into(pg_t child_pgid, PG *child, unsigned split_bits);
+  virtual void _split_into(pg_t child_pgid, PG *child, unsigned split_bits) = 0;
+
   loff_t get_log_write_pos() {
     return 0;
   }
@@ -795,6 +814,7 @@ public:
       epoch_start(0),
       block_writes(false), active(false), queue_snap_trim(false),
       waiting_on(0), errors(0), fixed(0), active_rep_scrub(0),
+      must_scrub(false), must_deep_scrub(false), must_repair(false),
       finalizing(false), is_chunky(false), state(INACTIVE),
       deep(false)
     {
@@ -816,6 +836,18 @@ public:
     ScrubMap primary_scrubmap;
     map<int,ScrubMap> received_maps;
     MOSDRepScrub *active_rep_scrub;
+    utime_t scrub_reg_stamp;  // stamp we registered for
+
+    // flags to indicate explicitly requested scrubs (by admin)
+    bool must_scrub, must_deep_scrub, must_repair;
+
+    // Maps from objects with erros to missing/inconsistent peers
+    map<hobject_t, set<int> > missing;
+    map<hobject_t, set<int> > inconsistent;
+    map<hobject_t, set<int> > inconsistent_snapcolls;
+
+    // Map from object with errors to good peer
+    map<hobject_t, pair<ScrubMap::object, int> > authoritative;
 
     // classic scrub
     bool finalizing;
@@ -840,6 +872,20 @@ public:
     // deep scrub
     bool deep;
 
+    list<Context*> callbacks;
+    void add_callback(Context *context) {
+      callbacks.push_back(context);
+    }
+    void run_callbacks() {
+      list<Context*> to_run;
+      to_run.swap(callbacks);
+      for (list<Context*>::iterator i = to_run.begin();
+	   i != to_run.end();
+	   ++i) {
+	(*i)->complete(0);
+      }
+    }
+
     static const char *state_string(const PG::Scrubber::State& state) {
       const char *ret = NULL;
       switch( state )
@@ -858,6 +904,21 @@ public:
 
     bool is_chunky_scrub_active() const { return state != INACTIVE; }
 
+    // classic (non chunk) scrubs block all writes
+    // chunky scrubs only block writes to a range
+    bool write_blocked_by_scrub(const hobject_t &soid) {
+      if (!block_writes)
+	return false;
+
+      if (!is_chunky)
+	return true;
+
+      if (soid >= start && soid < end)
+	return true;
+
+      return false;
+    }
+
     // clear all state
     void reset() {
       finalizing = false;
@@ -872,6 +933,10 @@ public:
       }
       received_maps.clear();
 
+      must_scrub = false;
+      must_deep_scrub = false;
+      must_repair = false;
+
       state = PG::Scrubber::INACTIVE;
       start = hobject_t();
       end = hobject_t();
@@ -879,13 +944,22 @@ public:
       errors = 0;
       fixed = 0;
       deep = false;
+      run_callbacks();
+      inconsistent.clear();
+      missing.clear();
+      authoritative.clear();
     }
 
   } scrubber;
 
+  bool scrub_after_recovery;
+
   int active_pushes;
 
   void repair_object(const hobject_t& soid, ScrubMap::object *po, int bad_peer, int ok_peer);
+  map<int, ScrubMap *>::const_iterator _select_auth_object(
+    const hobject_t &obj,
+    const map<int,ScrubMap*> &maps);
   bool _compare_scrub_objects(ScrubMap::object &auth,
 			      ScrubMap::object &candidate,
 			      ostream &errorstream);
@@ -893,11 +967,13 @@ public:
 			  map<hobject_t, set<int> > &missing,
 			  map<hobject_t, set<int> > &inconsistent,
 			  map<hobject_t, int> &authoritative,
+			  map<hobject_t, set<int> > &inconsistent_snapcolls,
 			  ostream &errorstream);
   void scrub();
   void classic_scrub();
   void chunky_scrub();
   void scrub_compare_maps();
+  void scrub_process_inconsistent();
   void scrub_finalize();
   void scrub_finish();
   void scrub_clear_state();
@@ -915,11 +991,24 @@ public:
   virtual void _scrub_finish() { }
   virtual coll_t get_temp_coll() = 0;
   virtual bool have_temp_coll() = 0;
+  virtual bool _report_snap_collection_errors(
+    const hobject_t &hoid,
+    int osd,
+    const map<string, bufferptr> &attrs,
+    const set<snapid_t> &snapcolls,
+    uint32_t nlinks,
+    ostream &out) { return false; };
+  virtual void check_snap_collections(
+    ino_t hino, const hobject_t &hoid,
+    const map<string, bufferptr> &attrs,
+    set<snapid_t> *snapcolls) {};
   void clear_scrub_reserved();
   void scrub_reserve_replicas();
   void scrub_unreserve_replicas();
   bool scrub_all_replicas_reserved() const;
   bool sched_scrub();
+  void reg_next_scrub();
+  void unreg_next_scrub();
 
   void replica_scrub(class MOSDRepScrub *op);
   void sub_op_scrub_map(OpRequestRef op);
@@ -962,7 +1051,7 @@ public:
     Formatter *f;
     QueryState(Formatter *f) : f(f) {}
     void print(std::ostream *out) const {
-      *out << "Query" << std::endl;
+      *out << "Query";
     }
   };
 
@@ -973,9 +1062,7 @@ public:
     MInfoRec(int from, pg_info_t &info, epoch_t msg_epoch) :
       from(from), info(info), msg_epoch(msg_epoch) {}
     void print(std::ostream *out) const {
-      *out << "MInfoRec from " << from
-	   << " info: " << info
-	   << std::endl;
+      *out << "MInfoRec from " << from << " info: " << info;
     }
   };
 
@@ -985,8 +1072,7 @@ public:
     MLogRec(int from, MOSDPGLog *msg) :
       from(from), msg(msg) {}
     void print(std::ostream *out) const {
-      *out << "MLogRec from " << from
-	   << std::endl;
+      *out << "MLogRec from " << from;
     }
   };
 
@@ -996,9 +1082,7 @@ public:
     MNotifyRec(int from, pg_notify_t &notify) :
       from(from), notify(notify) {}
     void print(std::ostream *out) const {
-      *out << "MNotifyRec from " << from
-	   << " notify: " << notify
-	   << std::endl;
+      *out << "MNotifyRec from " << from << " notify: " << notify;
     }
   };
 
@@ -1011,8 +1095,7 @@ public:
     void print(std::ostream *out) const {
       *out << "MQuery from " << from
 	   << " query_epoch " << query_epoch
-	   << " query: " << query
-	   << std::endl;
+	   << " query: " << query;
     }
   };
 
@@ -1023,20 +1106,14 @@ public:
     AdvMap(OSDMapRef osdmap, OSDMapRef lastmap, vector<int>& newup, vector<int>& newacting):
       osdmap(osdmap), lastmap(lastmap), newup(newup), newacting(newacting) {}
     void print(std::ostream *out) const {
-      *out << "AdvMap" << std::endl;
+      *out << "AdvMap";
     }
   };
 
-  struct RecoveryComplete : boost::statechart::event< RecoveryComplete > {
-    RecoveryComplete() : boost::statechart::event< RecoveryComplete >() {}
-    void print(std::ostream *out) const {
-      *out << "RecoveryComplete" << std::endl;
-    }
-  };
   struct ActMap : boost::statechart::event< ActMap > {
     ActMap() : boost::statechart::event< ActMap >() {}
     void print(std::ostream *out) const {
-      *out << "ActMap" << std::endl;
+      *out << "ActMap";
     }
   };
   struct Activate : boost::statechart::event< Activate > {
@@ -1044,13 +1121,13 @@ public:
     Activate(epoch_t q) : boost::statechart::event< Activate >(),
 			  query_epoch(q) {}
     void print(std::ostream *out) const {
-      *out << "Activate from " << query_epoch << std::endl;
+      *out << "Activate from " << query_epoch;
     }
   };
 #define TrivialEvent(T) struct T : boost::statechart::event< T > { \
     T() : boost::statechart::event< T >() {}			   \
     void print(std::ostream *out) const {			   \
-      *out << #T << std::endl;					   \
+      *out << #T;						   \
     }								   \
   };
   TrivialEvent(Initialize)
@@ -1065,6 +1142,19 @@ public:
   TrivialEvent(RemoteBackfillReserved)
   TrivialEvent(RemoteReservationRejected)
   TrivialEvent(RequestBackfill)
+  TrivialEvent(RequestRecovery)
+  TrivialEvent(RecoveryDone)
+
+  TrivialEvent(AllReplicasRecovered)
+  TrivialEvent(DoRecovery)
+  TrivialEvent(LocalRecoveryReserved)
+  TrivialEvent(RemoteRecoveryReserved)
+  TrivialEvent(AllRemotesReserved)
+  TrivialEvent(Recovering)
+  TrivialEvent(WaitRemoteBackfillReserved)
+  TrivialEvent(GoClean)
+
+  TrivialEvent(AllReplicasActivated)
 
   /* Encapsulates PG recovery process */
   class RecoveryState {
@@ -1292,11 +1382,15 @@ public:
       boost::statechart::result react(const AdvMap &advmap);
     };
 
-    struct NotBackfilling;
-    struct Active : boost::statechart::state< Active, Primary, NotBackfilling >, NamedState {
+    struct WaitLocalRecoveryReserved;
+    struct Activating;
+    struct Active : boost::statechart::state< Active, Primary, Activating >, NamedState {
       Active(my_context ctx);
       void exit();
 
+      const set<int> sorted_acting_set;
+      bool all_replicas_activated;
+
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< ActMap >,
@@ -1304,8 +1398,8 @@ public:
 	boost::statechart::custom_reaction< MInfoRec >,
 	boost::statechart::custom_reaction< MNotifyRec >,
 	boost::statechart::custom_reaction< MLogRec >,
-	boost::statechart::custom_reaction< RecoveryComplete >,
-	boost::statechart::custom_reaction< Backfilled >
+	boost::statechart::custom_reaction< Backfilled >,
+	boost::statechart::custom_reaction< AllReplicasActivated >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const ActMap&);
@@ -1316,12 +1410,33 @@ public:
       boost::statechart::result react(const Backfilled&) {
 	return discard_event();
       }
-      boost::statechart::result react(const RecoveryComplete&);
+      boost::statechart::result react(const AllReplicasActivated&);
+    };
+
+    struct Clean : boost::statechart::state< Clean, Active >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< DoRecovery, WaitLocalRecoveryReserved >
+      > reactions;
+      Clean(my_context ctx);
+      void exit();
+    };
+
+    struct Recovered : boost::statechart::state< Recovered, Active >, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< GoClean, Clean >,
+	boost::statechart::custom_reaction< AllReplicasActivated >
+      > reactions;
+      Recovered(my_context ctx);
+      void exit();
+      boost::statechart::result react(const AllReplicasActivated&) {
+	post_event(GoClean());
+	return forward_event();
+      }
     };
 
     struct Backfilling : boost::statechart::state< Backfilling, Active >, NamedState {
       typedef boost::mpl::list<
-	boost::statechart::transition< Backfilled, NotBackfilling >
+	boost::statechart::transition< Backfilled, Recovered >
 	> reactions;
       Backfilling(my_context ctx);
       void exit();
@@ -1354,8 +1469,8 @@ public:
       void exit();
     };
 
-    struct RepNotBackfilling;
-    struct ReplicaActive : boost::statechart::state< ReplicaActive, Started, RepNotBackfilling >, NamedState {
+    struct RepNotRecovering;
+    struct ReplicaActive : boost::statechart::state< ReplicaActive, Started, RepNotRecovering >, NamedState {
       ReplicaActive(my_context ctx);
       void exit();
 
@@ -1375,11 +1490,11 @@ public:
       boost::statechart::result react(const Activate&);
     };
 
-    struct RepBackfilling : boost::statechart::state< RepBackfilling, ReplicaActive >, NamedState {
+    struct RepRecovering : boost::statechart::state< RepRecovering, ReplicaActive >, NamedState {
       typedef boost::mpl::list<
-	boost::statechart::transition< Backfilled, RepNotBackfilling >
+	boost::statechart::transition< RecoveryDone, RepNotRecovering >
 	> reactions;
-      RepBackfilling(my_context ctx);
+      RepRecovering(my_context ctx);
       void exit();
     };
 
@@ -1394,11 +1509,63 @@ public:
       boost::statechart::result react(const RemoteReservationRejected &evt);
     };
 
-    struct RepNotBackfilling : boost::statechart::state< RepNotBackfilling, ReplicaActive>, NamedState {
+    struct RepWaitRecoveryReserved : boost::statechart::state< RepWaitRecoveryReserved, ReplicaActive >, NamedState {
       typedef boost::mpl::list<
-	boost::statechart::transition< RequestBackfill, RepWaitBackfillReserved >
+	boost::statechart::custom_reaction< RemoteRecoveryReserved >
 	> reactions;
-      RepNotBackfilling(my_context ctx);
+      RepWaitRecoveryReserved(my_context ctx);
+      void exit();
+      boost::statechart::result react(const RemoteRecoveryReserved &evt);
+    };
+
+    struct RepNotRecovering : boost::statechart::state< RepNotRecovering, ReplicaActive>, NamedState {
+      typedef boost::mpl::list<
+	boost::statechart::transition< RequestBackfill, RepWaitBackfillReserved >,
+        boost::statechart::transition< RequestRecovery, RepWaitRecoveryReserved >,
+	boost::statechart::transition< RecoveryDone, RepNotRecovering >  // for compat with pre-reservation peers
+	> reactions;
+      RepNotRecovering(my_context ctx);
+      void exit();
+    };
+
+    struct Recovering : boost::statechart::state< Recovering, Active >, NamedState {
+      typedef boost::mpl::list <
+	boost::statechart::custom_reaction< AllReplicasRecovered >,
+	boost::statechart::custom_reaction< RequestBackfill >
+	> reactions;
+      Recovering(my_context ctx);
+      void exit();
+      void release_reservations();
+      boost::statechart::result react(const AllReplicasRecovered &evt);
+      boost::statechart::result react(const RequestBackfill &evt);
+    };
+
+    struct WaitRemoteRecoveryReserved : boost::statechart::state< WaitRemoteRecoveryReserved, Active >, NamedState {
+      typedef boost::mpl::list <
+	boost::statechart::custom_reaction< RemoteRecoveryReserved >,
+	boost::statechart::transition< AllRemotesReserved, Recovering >
+	> reactions;
+      set<int>::const_iterator acting_osd_it;
+      WaitRemoteRecoveryReserved(my_context ctx);
+      boost::statechart::result react(const RemoteRecoveryReserved &evt);
+      void exit();
+    };
+
+    struct WaitLocalRecoveryReserved : boost::statechart::state< WaitLocalRecoveryReserved, Active >, NamedState {
+      typedef boost::mpl::list <
+	boost::statechart::transition< LocalRecoveryReserved, WaitRemoteRecoveryReserved >
+	> reactions;
+      WaitLocalRecoveryReserved(my_context ctx);
+      void exit();
+    };
+
+    struct Activating : boost::statechart::state< Activating, Active >, NamedState {
+      typedef boost::mpl::list <
+	boost::statechart::transition< AllReplicasRecovered, Recovered >,
+	boost::statechart::transition< DoRecovery, WaitLocalRecoveryReserved >,
+	boost::statechart::transition< RequestBackfill, WaitLocalBackfillReserved >
+	> reactions;
+      Activating(my_context ctx);
       void exit();
     };
 
@@ -1412,12 +1579,16 @@ public:
 	boost::statechart::custom_reaction< MQuery >,
 	boost::statechart::custom_reaction< MLogRec >,
 	boost::statechart::custom_reaction< MInfoRec >,
-	boost::statechart::custom_reaction< ActMap >
+	boost::statechart::custom_reaction< ActMap >,
+	boost::statechart::custom_reaction< RecoveryDone >
 	> reactions;
       boost::statechart::result react(const MQuery& query);
       boost::statechart::result react(const MLogRec& logevt);
       boost::statechart::result react(const MInfoRec& infoevt);
       boost::statechart::result react(const ActMap&);
+      boost::statechart::result react(const RecoveryDone&) {
+	return discard_event();
+      }
     };
 
     struct GetLog;
@@ -1510,7 +1681,11 @@ public:
     };
 
     struct Incomplete : boost::statechart::state< Incomplete, Peering>, NamedState {
+      typedef boost::mpl::list <
+	boost::statechart::custom_reaction< AdvMap >
+	> reactions;
       Incomplete(my_context ctx);
+      boost::statechart::result react(const AdvMap &advmap);
       void exit();
     };
 
@@ -1606,7 +1781,9 @@ public:
   void trim_peers();
 
   std::string get_corrupt_pg_log_name() const;
-  void read_state(ObjectStore *store);
+  void read_state(ObjectStore *store, bufferlist &bl);
+  static epoch_t peek_map_epoch(ObjectStore *store,
+				coll_t coll, bufferlist *bl);
   coll_t make_snap_collection(ObjectStore::Transaction& t, snapid_t sn);
   void update_snap_collections(vector<pg_log_entry_t> &log_entries,
 			       ObjectStore::Transaction& t);
@@ -1635,6 +1812,7 @@ public:
   void fulfill_info(int from, const pg_query_t &query, 
 		    pair<int, pg_info_t> &notify_info);
   void fulfill_log(int from, const pg_query_t &query, epoch_t query_epoch);
+  bool is_split(OSDMapRef lastmap, OSDMapRef nextmap);
   bool acting_up_affected(const vector<int>& newup, const vector<int>& newacting);
 
   // OpRequest queueing
@@ -1645,7 +1823,8 @@ public:
   bool can_discard_request(OpRequestRef op);
 
   bool must_delay_request(OpRequestRef op);
-  void queue_op(OpRequestRef op);
+
+  static bool split_request(OpRequestRef op, unsigned match, unsigned bits);
 
   bool old_peering_msg(epoch_t reply_epoch, epoch_t query_epoch);
   bool old_peering_evt(CephPeeringEvtRef evt) {
@@ -1676,7 +1855,6 @@ public:
 			  vector<int>& newup, vector<int>& newacting,
 			  RecoveryCtx *rctx);
   void handle_activate_map(RecoveryCtx *rctx);
-  void handle_recovery_complete(RecoveryCtx *rctx);
   void handle_create(RecoveryCtx *rctx);
   void handle_loaded(RecoveryCtx *rctx);
   void handle_query_state(Formatter *f);
@@ -1724,4 +1902,6 @@ ostream& operator<<(ostream& out, const PG& pg);
 void intrusive_ptr_add_ref(PG *pg);
 void intrusive_ptr_release(PG *pg);
 
+typedef boost::intrusive_ptr<PG> PGRef;
+
 #endif
