@@ -12,7 +12,7 @@
 #include "common/Thread.h"
 
 #include "Objecter.h"
-#include "Filer.h"
+#include "Striper.h"
 
 class CephContext;
 class WritebackHandler;
@@ -84,10 +84,12 @@ class ObjectCacher {
     // states
     static const int STATE_MISSING = 0;
     static const int STATE_CLEAN = 1;
-    static const int STATE_DIRTY = 2;
-    static const int STATE_RX = 3;
-    static const int STATE_TX = 4;
-    
+    static const int STATE_ZERO = 2;   // NOTE: these are *clean* zeros
+    static const int STATE_DIRTY = 3;
+    static const int STATE_RX = 4;
+    static const int STATE_TX = 5;
+    static const int STATE_ERROR = 6; // a read error occurred
+
   private:
     // my fields
     int state;
@@ -102,16 +104,19 @@ class ObjectCacher {
     tid_t last_write_tid;  // version of bh (if non-zero)
     utime_t last_write;
     SnapContext snapc;
+    int error; // holds return value for failed reads
     
     map< loff_t, list<Context*> > waitfor_read;
     
-  public:
     // cons
     BufferHead(Object *o) : 
       state(STATE_MISSING),
       ref(0),
       ob(o),
-      last_write_tid(0) {}
+      last_write_tid(0),
+      error(0) {
+      ex.start = ex.length = 0;
+    }
   
     // extent
     loff_t start() const { return ex.start; }
@@ -132,8 +137,10 @@ class ObjectCacher {
     bool is_missing() { return state == STATE_MISSING; }
     bool is_dirty() { return state == STATE_DIRTY; }
     bool is_clean() { return state == STATE_CLEAN; }
+    bool is_zero() { return state == STATE_ZERO; }
     bool is_tx() { return state == STATE_TX; }
     bool is_rx() { return state == STATE_RX; }
+    bool is_error() { return state == STATE_ERROR; }
     
     // reference counting
     int get() {
@@ -148,111 +155,145 @@ class ObjectCacher {
       return ref;
     }
   };
+
+  // ******* Object *********
+  class Object : public LRUObject {
+  private:
+    // ObjectCacher::Object fields
+    int ref;
+    ObjectCacher *oc;
+    sobject_t oid;
+    friend class ObjectSet;
+
+  public:
+    ObjectSet *oset;
+    xlist<Object*>::item set_item;
+    object_locator_t oloc;
+    
+    bool complete;
+    bool exists;
+
+  public:
+    map<loff_t, BufferHead*>     data;
+
+    tid_t last_write_tid;  // version of bh (if non-zero)
+    tid_t last_commit_tid; // last update commited.
+
+    int dirty_or_tx;
+
+    map< tid_t, list<Context*> > waitfor_commit;
+
+  public:
+    Object(const Object& other);
+    const Object& operator=(const Object& other);
+
+    Object(ObjectCacher *_oc, sobject_t o, ObjectSet *os, object_locator_t& l) : 
+      ref(0),
+      oc(_oc),
+      oid(o), oset(os), set_item(this), oloc(l),
+      complete(false), exists(true),
+      last_write_tid(0), last_commit_tid(0),
+      dirty_or_tx(0) {
+      // add to set
+      os->objects.push_back(&set_item);
+    }
+    ~Object() {
+      assert(ref == 0);
+      assert(data.empty());
+      assert(dirty_or_tx == 0);
+      set_item.remove_myself();
+    }
+
+    sobject_t get_soid() { return oid; }
+    object_t get_oid() { return oid.oid; }
+    snapid_t get_snap() { return oid.snap; }
+    ObjectSet *get_object_set() { return oset; }
+    
+    object_locator_t& get_oloc() { return oloc; }
+    void set_object_locator(object_locator_t& l) { oloc = l; }
+
+    bool can_close() {
+      if (lru_is_expireable()) {
+	assert(data.empty());
+	assert(waitfor_commit.empty());
+	return true;
+      }
+      return false;
+    }
+
+    /**
+     * Check buffers and waiters for consistency
+     * - no overlapping buffers
+     * - index in map matches BH
+     * - waiters fall within BH
+     */
+    void audit_buffers();
+
+    /**
+     * find first buffer that includes or follows an offset
+     *
+     * @param offset object byte offset
+     * @return iterator pointing to buffer, or data.end()
+     */
+    map<loff_t,BufferHead*>::iterator data_lower_bound(loff_t offset) {
+      map<loff_t,BufferHead*>::iterator p = data.lower_bound(offset);
+      if (p != data.begin() &&
+	  (p == data.end() || p->first > offset)) {
+	p--;     // might overlap!
+	if (p->first + p->second->length() <= offset)
+	  p++;   // doesn't overlap.
+      }
+      return p;
+    }
+
+    // bh
+    // add to my map
+    void add_bh(BufferHead *bh) {
+      if (data.empty())
+	get();
+      assert(data.count(bh->start()) == 0);
+      data[bh->start()] = bh;
+    }
+    void remove_bh(BufferHead *bh) {
+      assert(data.count(bh->start()));
+      data.erase(bh->start());
+      if (data.empty())
+	put();
+    }
+
+    bool is_empty() { return data.empty(); }
+
+    // mid-level
+    BufferHead *split(BufferHead *bh, loff_t off);
+    void merge_left(BufferHead *left, BufferHead *right);
+    void try_merge_bh(BufferHead *bh);
+
+    bool is_cached(loff_t off, loff_t len);
+    int map_read(OSDRead *rd,
+                 map<loff_t, BufferHead*>& hits,
+                 map<loff_t, BufferHead*>& missing,
+                 map<loff_t, BufferHead*>& rx,
+		 map<loff_t, BufferHead*>& errors);
+    BufferHead *map_write(OSDWrite *wr);
+    
+    void truncate(loff_t s);
+    void discard(loff_t off, loff_t len);
+
+    // reference counting
+    int get() {
+      assert(ref >= 0);
+      if (ref == 0) lru_pin();
+      return ++ref;
+    }
+    int put() {
+      assert(ref > 0);
+      if (ref == 1) lru_unpin();
+      --ref;
+      return ref;
+    }
+  };
   
 
-  // ******* Object *********
-  class Object {
-  private:
-    // ObjectCacher::Object fields
-    ObjectCacher *oc;
-    sobject_t oid;
-    friend class ObjectSet;
-
-  public:
-    ObjectSet *oset;
-    xlist<Object*>::item set_item;
-    object_locator_t oloc;
-    
-
-  public:
-    map<loff_t, BufferHead*>     data;
-
-    tid_t last_write_tid;  // version of bh (if non-zero)
-    tid_t last_commit_tid; // last update commited.
-
-    int dirty_or_tx;
-
-    map< tid_t, list<Context*> > waitfor_commit;
-    list<Context*> waitfor_rd;
-    list<Context*> waitfor_wr;
-
-    // lock
-    static const int LOCK_NONE = 0;
-    static const int LOCK_WRLOCKING = 1;
-    static const int LOCK_WRLOCK = 2;
-    static const int LOCK_WRUNLOCKING = 3;
-    static const int LOCK_RDLOCKING = 4;
-    static const int LOCK_RDLOCK = 5;
-    static const int LOCK_RDUNLOCKING = 6;
-    static const int LOCK_UPGRADING = 7;    // rd -> wr
-    static const int LOCK_DOWNGRADING = 8;  // wr -> rd
-    int lock_state;
-    int wrlock_ref;  // how many ppl want or are using a WRITE lock
-    int rdlock_ref;  // how many ppl want or are using a READ lock
-
-  public:
-    Object(const Object& other);
-    const Object& operator=(const Object& other);
-
-    Object(ObjectCacher *_oc, sobject_t o, ObjectSet *os, object_locator_t& l) : 
-      oc(_oc),
-      oid(o), oset(os), set_item(this), oloc(l),
-      last_write_tid(0), last_commit_tid(0),
-      dirty_or_tx(0),
-      lock_state(LOCK_NONE), wrlock_ref(0), rdlock_ref(0) {
-      // add to set
-      os->objects.push_back(&set_item);
-    }
-    ~Object() {
-      assert(data.empty());
-      assert(dirty_or_tx == 0);
-      set_item.remove_myself();
-    }
-
-    sobject_t get_soid() { return oid; }
-    object_t get_oid() { return oid.oid; }
-    snapid_t get_snap() { return oid.snap; }
-    ObjectSet *get_object_set() { return oset; }
-    
-    object_locator_t& get_oloc() { return oloc; }
-    void set_object_locator(object_locator_t& l) { oloc = l; }
-
-    bool can_close() {
-      return data.empty() && lock_state == LOCK_NONE &&
-        waitfor_commit.empty() &&
-        waitfor_rd.empty() && waitfor_wr.empty() &&
-	dirty_or_tx == 0;
-    }
-
-    // bh
-    // add to my map
-    void add_bh(BufferHead *bh) {
-      assert(data.count(bh->start()) == 0);
-      data[bh->start()] = bh;
-    }
-    void remove_bh(BufferHead *bh) {
-      assert(data.count(bh->start()));
-      data.erase(bh->start());
-    }
-    bool is_empty() { return data.empty(); }
-
-    // mid-level
-    BufferHead *split(BufferHead *bh, loff_t off);
-    void merge_left(BufferHead *left, BufferHead *right);
-    void try_merge_bh(BufferHead *bh);
-
-    bool is_cached(loff_t off, loff_t len);
-    int map_read(OSDRead *rd,
-                 map<loff_t, BufferHead*>& hits,
-                 map<loff_t, BufferHead*>& missing,
-                 map<loff_t, BufferHead*>& rx);
-    BufferHead *map_write(OSDWrite *wr);
-    
-    void truncate(loff_t s);
-    void discard(loff_t off, loff_t len);
-  };
-  
-
   struct ObjectSet {
     void *parent;
 
@@ -263,10 +304,13 @@ class ObjectCacher {
     xlist<Object*> objects;
 
     int dirty_or_tx;
+    bool return_enoent;
 
     ObjectSet(void *p, int64_t _poolid, inodeno_t i)
       : parent(p), ino(i), truncate_seq(0),
-	truncate_size(0), poolid(_poolid), dirty_or_tx(0) {}
+	truncate_size(0), poolid(_poolid), dirty_or_tx(0),
+	return_enoent(false) {}
+
   };
 
 
@@ -278,13 +322,17 @@ class ObjectCacher {
   string name;
   Mutex& lock;
   
+  int64_t max_dirty, target_dirty, max_size, max_objects;
+  utime_t max_dirty_age;
+
   flush_set_callback_t flush_set_callback;
   void *flush_set_callback_arg;
 
   vector<hash_map<sobject_t, Object*> > objects; // indexed by pool_id
 
   set<BufferHead*>    dirty_bh;
-  LRU   lru_dirty, lru_rest;
+  LRU   bh_lru_dirty, bh_lru_rest;
+  LRU   ob_lru;
 
   Cond flusher_cond;
   bool flusher_stop;
@@ -315,13 +363,15 @@ class ObjectCacher {
 
   // bh stats
   Cond  stat_cond;
-  int   stat_waiter;
 
   loff_t stat_clean;
+  loff_t stat_zero;
   loff_t stat_dirty;
   loff_t stat_rx;
   loff_t stat_tx;
   loff_t stat_missing;
+  loff_t stat_error;
+  loff_t stat_dirty_waiting;   // bytes that writers are waiting on to write
 
   void verify_stats() const;
 
@@ -330,13 +380,19 @@ class ObjectCacher {
   loff_t get_stat_tx() { return stat_tx; }
   loff_t get_stat_rx() { return stat_rx; }
   loff_t get_stat_dirty() { return stat_dirty; }
+  loff_t get_stat_dirty_waiting() { return stat_dirty_waiting; }
   loff_t get_stat_clean() { return stat_clean; }
+  loff_t get_stat_zero() { return stat_zero; }
 
   void touch_bh(BufferHead *bh) {
     if (bh->is_dirty())
-      lru_dirty.lru_touch(bh);
+      bh_lru_dirty.lru_touch(bh);
     else
-      lru_rest.lru_touch(bh);
+      bh_lru_rest.lru_touch(bh);
+    touch_ob(bh->ob);
+  }
+  void touch_ob(Object *ob) {
+    ob_lru.lru_touch(ob);
   }
 
   // bh states
@@ -347,11 +403,13 @@ class ObjectCacher {
   
   void mark_missing(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_MISSING); };
   void mark_clean(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_CLEAN); };
+  void mark_zero(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_ZERO); };
   void mark_rx(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_RX); };
   void mark_tx(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_TX); };
+  void mark_error(BufferHead *bh) { bh_set_state(bh, BufferHead::STATE_ERROR); };
   void mark_dirty(BufferHead *bh) { 
     bh_set_state(bh, BufferHead::STATE_DIRTY); 
-    lru_dirty.lru_touch(bh);
+    bh_lru_dirty.lru_touch(bh);
     //bh->set_dirty_stamp(ceph_clock_now(g_ceph_context));
   };
 
@@ -362,24 +420,32 @@ class ObjectCacher {
   void bh_read(BufferHead *bh);
   void bh_write(BufferHead *bh);
 
-  void trim(loff_t max=-1);
+  void trim(loff_t max_bytes=-1, loff_t max_objects=-1);
   void flush(loff_t amount=0);
 
-  bool flush(Object *o);
+  /**
+   * flush a range of buffers
+   *
+   * Flush any buffers that intersect the specified extent.  If len==0,
+   * flush *all* buffers for the object.
+   *
+   * @param o object
+   * @param off start offset
+   * @param len extent length, or 0 for entire object
+   * @return true if object was already clean/flushed.
+   */
+  bool flush(Object *o, loff_t off, loff_t len);
   loff_t release(Object *o);
   void purge(Object *o);
 
-  void rdlock(Object *o);
-  void rdunlock(Object *o);
-  void wrlock(Object *o);
-  void wrunlock(Object *o);
+  int _readx(OSDRead *rd, ObjectSet *oset, Context *onfinish,
+	     bool external_call);
 
  public:
   void bh_read_finish(int64_t poolid, sobject_t oid, loff_t offset,
 		      uint64_t length, bufferlist &bl, int r);
   void bh_write_commit(int64_t poolid, sobject_t oid, loff_t offset,
 		       uint64_t length, tid_t t, int r);
-  void lock_ack(int64_t poolid, list<sobject_t>& oids, tid_t tid);
 
   class C_ReadFinish : public Context {
     ObjectCacher *oc;
@@ -405,26 +471,12 @@ class ObjectCacher {
   public:
     tid_t tid;
     C_WriteCommit(ObjectCacher *c, int64_t _poolid, sobject_t o, loff_t s, uint64_t l) :
-      oc(c), poolid(_poolid), oid(o), start(s), length(l) {}
+      oc(c), poolid(_poolid), oid(o), start(s), length(l), tid(0) {}
     void finish(int r) {
       oc->bh_write_commit(poolid, oid, start, length, tid, r);
     }
   };
 
-  class C_LockAck : public Context {
-    ObjectCacher *oc;
-  public:
-    int64_t poolid;
-    list<sobject_t> oids;
-    tid_t tid;
-    C_LockAck(ObjectCacher *c, int64_t _poolid, sobject_t o) : oc(c), poolid(_poolid) {
-      oids.push_back(o);
-    }
-    void finish(int r) {
-      oc->lock_ack(poolid, oids, tid);
-    }
-  };
-
   void perf_start();
   void perf_stop();
 
@@ -432,7 +484,9 @@ class ObjectCacher {
 
   ObjectCacher(CephContext *cct_, string name, WritebackHandler& wb, Mutex& l,
 	       flush_set_callback_t flush_callback,
-	       void *flush_callback_arg);
+	       void *flush_callback_arg,
+	       uint64_t max_bytes, uint64_t max_objects,
+	       uint64_t max_dirty, uint64_t target_dirty, double max_age);
   ~ObjectCacher();
 
   void start() {
@@ -455,11 +509,15 @@ class ObjectCacher {
     Context *onfinish;
   public:
     C_RetryRead(ObjectCacher *_oc, OSDRead *r, ObjectSet *os, Context *c) : oc(_oc), rd(r), oset(os), onfinish(c) {}
-    void finish(int) {
-      int r = oc->readx(rd, oset, onfinish);
-      if (r > 0 && onfinish) {
-        onfinish->finish(r);
-        delete onfinish;
+    void finish(int r) {
+      if (r < 0) {
+	if (onfinish)
+	  onfinish->complete(r);
+	return;
+      }
+      int ret = oc->_readx(rd, oset, onfinish, false);
+      if (ret != 0 && onfinish) {
+        onfinish->complete(ret);
       }
     }
   };
@@ -467,17 +525,25 @@ class ObjectCacher {
 
 
   // non-blocking.  async.
+
+  /**
+   * @note total read size must be <= INT_MAX, since
+   * the return value is total bytes read
+   */
   int readx(OSDRead *rd, ObjectSet *oset, Context *onfinish);
-  int writex(OSDWrite *wr, ObjectSet *oset);
+  int writex(OSDWrite *wr, ObjectSet *oset, Mutex& wait_on_lock);
   bool is_cached(ObjectSet *oset, vector<ObjectExtent>& extents, snapid_t snapid);
 
+private:
   // write blocking
-  bool wait_for_write(uint64_t len, Mutex& lock);
+  int _wait_for_write(OSDWrite *wr, uint64_t len, ObjectSet *oset, Mutex& lock);
   
+public:
   bool set_is_cached(ObjectSet *oset);
   bool set_is_dirty_or_committing(ObjectSet *oset);
 
   bool flush_set(ObjectSet *oset, Context *onfinish=0);
+  bool flush_set(ObjectSet *oset, vector<ObjectExtent>& ex, Context *onfinish=0);
   void flush_all(Context *onfinish=0);
 
   bool commit_set(ObjectSet *oset, Context *oncommit);
@@ -489,13 +555,31 @@ class ObjectCacher {
 
   void discard_set(ObjectSet *oset, vector<ObjectExtent>& ex);
 
+  // cache sizes
+  void set_max_dirty(int64_t v) {
+    max_dirty = v;
+  }
+  void set_target_dirty(int64_t v) {
+    target_dirty = v;
+  }
+  void set_max_size(int64_t v) {
+    max_size = v;
+  }
+  void set_max_dirty_age(double a) {
+    max_dirty_age.set_from_double(a);
+  }
+  void set_max_objects(int64_t v) {
+    max_objects = v;
+  }
+
+
   // file functions
 
   /*** async+caching (non-blocking) file interface ***/
   int file_is_cached(ObjectSet *oset, ceph_file_layout *layout, snapid_t snapid,
 		     loff_t offset, uint64_t len) {
     vector<ObjectExtent> extents;
-    Filer::file_to_extents(cct, oset->ino, layout, offset, len, extents);
+    Striper::file_to_extents(cct, oset->ino, layout, offset, len, extents);
     return is_cached(oset, extents, snapid);
   }
 
@@ -505,23 +589,24 @@ class ObjectCacher {
 		int flags,
                 Context *onfinish) {
     OSDRead *rd = prepare_read(snapid, bl, flags);
-    Filer::file_to_extents(cct, oset->ino, layout, offset, len, rd->extents);
+    Striper::file_to_extents(cct, oset->ino, layout, offset, len, rd->extents);
     return readx(rd, oset, onfinish);
   }
 
   int file_write(ObjectSet *oset, ceph_file_layout *layout, const SnapContext& snapc,
                  loff_t offset, uint64_t len, 
-                 bufferlist& bl, utime_t mtime, int flags) {
+                 bufferlist& bl, utime_t mtime, int flags,
+		 Mutex& wait_on_lock) {
     OSDWrite *wr = prepare_write(snapc, bl, mtime, flags);
-    Filer::file_to_extents(cct, oset->ino, layout, offset, len, wr->extents);
-    return writex(wr, oset);
+    Striper::file_to_extents(cct, oset->ino, layout, offset, len, wr->extents);
+    return writex(wr, oset, wait_on_lock);
   }
 };
 
 
 inline ostream& operator<<(ostream& out, ObjectCacher::BufferHead &bh)
 {
-  out << "bh["
+  out << "bh[ " << &bh << " "
       << bh.start() << "~" << bh.length()
       << " " << bh.ob
       << " (" << bh.bl.length() << ")"
@@ -530,9 +615,22 @@ inline ostream& operator<<(ostream& out, ObjectCacher::BufferHead &bh)
   if (bh.is_rx()) out << " rx";
   if (bh.is_dirty()) out << " dirty";
   if (bh.is_clean()) out << " clean";
+  if (bh.is_zero()) out << " zero";
   if (bh.is_missing()) out << " missing";
   if (bh.bl.length() > 0) out << " firstbyte=" << (int)bh.bl[0];
+  if (bh.error) out << " error=" << bh.error;
   out << "]";
+  out << " waiters = {";
+  for (map<loff_t, list<Context*> >::const_iterator it = bh.waitfor_read.begin();
+       it != bh.waitfor_read.end(); ++it) {
+    out << " " << it->first << "->[";
+    for (list<Context*>::const_iterator lit = it->second.begin();
+	 lit != it->second.end(); ++lit) {
+	 out << *lit << ", ";
+    }
+    out << "]";
+  }
+  out << "}";
   return out;
 }
 
@@ -551,14 +649,10 @@ inline ostream& operator<<(ostream& out, ObjectCacher::Object &ob)
       << ob.get_soid() << " oset " << ob.oset << dec
       << " wr " << ob.last_write_tid << "/" << ob.last_commit_tid;
 
-  switch (ob.lock_state) {
-  case ObjectCacher::Object::LOCK_WRLOCKING: out << " wrlocking"; break;
-  case ObjectCacher::Object::LOCK_WRLOCK: out << " wrlock"; break;
-  case ObjectCacher::Object::LOCK_WRUNLOCKING: out << " wrunlocking"; break;
-  case ObjectCacher::Object::LOCK_RDLOCKING: out << " rdlocking"; break;
-  case ObjectCacher::Object::LOCK_RDLOCK: out << " rdlock"; break;
-  case ObjectCacher::Object::LOCK_RDUNLOCKING: out << " rdunlocking"; break;
-  }
+  if (ob.complete)
+    out << " COMPLETE";
+  if (!ob.exists)
+    out << " !EXISTS";
 
   out << "]";
   return out;
