@@ -155,22 +155,41 @@ class RBD(object):
         :param stripe_count: objects to stripe over before looping
         :type stripe_count: int
         :raises: :class:`ImageExists`
+        :raises: :class:`TypeError`
+        :raises: :class:`InvalidArgument`
+        :raises: :class:`FunctionNotSupported`
         """
         if order is None:
             order = 0
         if not isinstance(name, str):
             raise TypeError('name must be a string')
         if old_format:
+            if features != 0 or stripe_unit != 0 or stripe_count != 0:
+                raise InvalidArgument('format 1 images do not support feature'
+                                      ' masks or non-default striping')
             ret = self.librbd.rbd_create(ioctx.io, c_char_p(name),
                                          c_uint64(size),
                                          byref(c_int(order)))
         else:
-            ret = self.librbd.rbd_create3(ioctx.io, c_char_p(name),
-                                          c_uint64(size),
-                                          c_uint64(features),
-                                          byref(c_int(order)),
-                                          c_uint64(stripe_unit),
-                                          c_uint64(stripe_count))
+            if not hasattr(self.librbd, 'rbd_create2'):
+                raise FunctionNotSupported('installed version of librbd does'
+                                           ' not support format 2 images')
+            has_create3 = hasattr(self.librbd, 'rbd_create3')
+            if (stripe_unit != 0 or stripe_count != 0) and not has_create3:
+                raise FunctionNotSupported('installed version of librbd does'
+                                           ' not support stripe unit or count')
+            if has_create3:
+                ret = self.librbd.rbd_create3(ioctx.io, c_char_p(name),
+                                              c_uint64(size),
+                                              c_uint64(features),
+                                              byref(c_int(order)),
+                                              c_uint64(stripe_unit),
+                                              c_uint64(stripe_count))
+            else:
+                ret = self.librbd.rbd_create2(ioctx.io, c_char_p(name),
+                                              c_uint64(size),
+                                              c_uint64(features),
+                                              byref(c_int(order)))
         if ret < 0:
             raise make_ex(ret, 'error creating image')
 
@@ -282,10 +301,15 @@ class Image(object):
     if the image has been deleted.
     """
 
-    def __init__(self, ioctx, name, snapshot=None):
+    def __init__(self, ioctx, name, snapshot=None, read_only=False):
         """
         Open the image at the given snapshot.
-        If a snapshot is specified, the image will be read-only.
+        If a snapshot is specified, the image will be read-only, unless
+        :func:`Image.set_snap` is called later.
+
+        If read-only mode is used, metadata for the :class:`Image`
+        object (such as which snapshots exist) may become obsolete. See
+        the C api for more details.
 
         To clean up from opening the image, :func:`Image.close` should
         be called.  For ease of use, this is done automatically when
@@ -297,6 +321,8 @@ class Image(object):
         :type name: str
         :param snapshot: which snapshot to read from
         :type snaphshot: str
+        :param read_only: whether to open the image in read-only mode
+        :type read_only: bool
         """
         self.closed = True
         self.librbd = CDLL('librbd.so.1')
@@ -306,8 +332,16 @@ class Image(object):
             raise TypeError('name must be a string')
         if snapshot is not None and not isinstance(snapshot, str):
             raise TypeError('snapshot must be a string or None')
-        ret = self.librbd.rbd_open(ioctx.io, c_char_p(name),
-                                   byref(self.image), c_char_p(snapshot))
+        if read_only:
+            if not hasattr(self.librbd, 'rbd_open_read_only'):
+                raise FunctionNotSupported('installed version of librbd does '
+                                           'not support open in read-only mode')
+            ret = self.librbd.rbd_open_read_only(ioctx.io, c_char_p(name),
+                                                 byref(self.image),
+                                                 c_char_p(snapshot))
+        else:
+            ret = self.librbd.rbd_open(ioctx.io, c_char_p(name),
+                                       byref(self.image), c_char_p(snapshot))
         if ret != 0:
             raise make_ex(ret, 'error opening image %s at snapshot %s' % (name, snapshot))
         self.closed = False
