@@ -158,9 +158,17 @@ void RGWListBucket_ObjStore_SWIFT::send_response()
       if (name.compare(delimiter) == 0)
         goto next;
 
-      s->formatter->open_object_section("object");
-      s->formatter->dump_string("name", pref_iter->first);
-      s->formatter->close_section();
+        s->formatter->open_object_section("subdir");
+
+        /* swift is a bit inconsistent here */
+        switch (s->format) {
+          case RGW_FORMAT_XML:
+            s->formatter->dump_string("name", pref_iter->first);
+            break;
+          default:
+            s->formatter->dump_string("subdir", pref_iter->first);
+        }
+        s->formatter->close_section();
     }
 next:
     if (do_objs)
@@ -304,7 +312,10 @@ int RGWPutObj_ObjStore_SWIFT::get_params()
       suffix++;
       if (*suffix) {
         string suffix_str(suffix);
-        s->generic_attrs[RGW_ATTR_CONTENT_TYPE] = rgw_find_mime_by_ext(suffix_str);
+	const char *mime = rgw_find_mime_by_ext(suffix_str);
+	if (mime) {
+          s->generic_attrs[RGW_ATTR_CONTENT_TYPE] = mime;
+	}
       }
     }
   }
@@ -612,6 +623,13 @@ RGWOp *RGWHandler_ObjStore_Obj_SWIFT::op_copy()
 
 int RGWHandler_ObjStore_SWIFT::authorize()
 {
+  if (!s->os_auth_token) {
+    /* anonymous access */
+    rgw_get_anon_user(s->user);
+    s->perm_mask = RGW_PERM_FULL_CONTROL;
+    return 0;
+  }
+
   bool authorized = rgw_swift->verify_swift_token(store, s);
   if (!authorized)
     return -EPERM;
