@@ -19,6 +19,7 @@ TYPE(entity_addr_t)
 
 #include "osd/OSDMap.h"
 TYPE(osd_info_t)
+TYPE(osd_xinfo_t)
 TYPEWITHSTRAYDATA(OSDMap)
 TYPEWITHSTRAYDATA(OSDMap::Incremental)
 
@@ -35,12 +36,12 @@ TYPE(pg_t)
 TYPE(coll_t)
 TYPE(osd_stat_t)
 TYPE(OSDSuperblock)
-TYPE(pool_snap_info_t)
+TYPE_FEATUREFUL(pool_snap_info_t)
 TYPE_FEATUREFUL(pg_pool_t)
 TYPE(object_stat_sum_t)
 TYPE(object_stat_collection_t)
 TYPE(pg_stat_t)
-TYPE(pool_stat_t)
+TYPE_FEATUREFUL(pool_stat_t)
 TYPE(pg_history_t)
 TYPE(pg_info_t)
 TYPE(pg_interval_t)
@@ -106,13 +107,15 @@ TYPE(ObjectMetaInfo)
 TYPE(ObjectCacheInfo)
 TYPE(RGWCacheNotifyInfo)
 
-#include "rgw/rgw_cls_api.h"
+#include "cls/rgw/cls_rgw_types.h"
 TYPE(rgw_bucket_pending_info)
 TYPE(rgw_bucket_dir_entry_meta)
 TYPE(rgw_bucket_dir_entry)
 TYPE(rgw_bucket_category_stats)
 TYPE(rgw_bucket_dir_header)
 TYPE(rgw_bucket_dir)
+
+#include "cls/rgw/cls_rgw_ops.h"
 TYPE(rgw_cls_obj_prepare_op)
 TYPE(rgw_cls_obj_complete_op)
 TYPE(rgw_cls_list_op)
@@ -132,7 +135,7 @@ TYPE(rgw_obj)
 TYPE(rgw_log_entry)
 TYPE(rgw_intent_log_entry)
 
-#include "librbd/cls_rbd.h"
+#include "cls/rbd/cls_rbd.h"
 TYPE(cls_rbd_parent)
 TYPE(cls_rbd_snap)
 
