@@ -3,6 +3,11 @@
 # abort on failure
 set -e
 
+[ -z "$CEPH_NUM_MON" ] && CEPH_NUM_MON="$MON"
+[ -z "$CEPH_NUM_OSD" ] && CEPH_NUM_OSD="$OSD"
+[ -z "$CEPH_NUM_MDS" ] && CEPH_NUM_MDS="$MDS"
+[ -z "$CEPH_NUM_RGW" ] && CEPH_NUM_RGW="$RGW"
+
 [ -z "$CEPH_NUM_MON" ] && CEPH_NUM_MON=3
 [ -z "$CEPH_NUM_OSD" ] && CEPH_NUM_OSD=1
 [ -z "$CEPH_NUM_MDS" ] && CEPH_NUM_MDS=3
@@ -21,7 +26,7 @@ ip=""
 nodaemon=0
 smallmds=0
 overwrite_conf=1
-cephx=0
+cephx=1 #turn cephx on by default
 
 MON_ADDR=""
 
@@ -43,7 +48,8 @@ usage=$usage"\t--nodaemon: use ceph-run as wrapper for mon/osd/mds\n"
 usage=$usage"\t--smallmds: limit mds cache size\n"
 usage=$usage"\t-m ip:port\t\tspecify monitor address\n"
 usage=$usage"\t-k keep old configuration files\n"
-usage=$usage"\t-x disable authx\n"
+usage=$usage"\t-x enable cephx (on by default)\n"
+usage=$usage"\t-X disable cephx\n"
 usage=$usage"\t-o config\t\t add extra config parameters to mds section\n"
 
 usage_exit() {
@@ -117,7 +123,10 @@ case $1 in
 	    shift
 	    ;;
     -x )
-	    cephx=1
+	    cephx=1 # this is on be default, flag exists for historical consistency
+	    ;;
+    -X )
+	    cephx=0
 	    ;;
     -k )
 	    overwrite_conf=0
@@ -281,14 +290,25 @@ if [ "$start_mon" -eq 1 ]; then
 [global]
         osd pg bits = 3
         osd pgp bits = 5  ; (invalid, but ceph should cope!)
+        osd crush chooseleaf type = 0
+        osd pool default min size = 1
 EOF
-			[ "$cephx" -eq 1 ] && cat<<EOF >> $conf
+if [ "$cephx" -eq 1 ] ; then
+cat <<EOF >> $conf
         auth supported = cephx
 EOF
+else
+cat <<EOF >> $conf
+	auth cluster required = none
+	auth service required = none
+	auth client required = none
+EOF
+fi
 			cat <<EOF >> $conf
 
 [client]
         keyring = $keyring_fn
+        log file = out/\$name.log
 
 [mds]
 $DAEMONOPTS
